from __future__ import annotations

import numpy as np
import xarray as xr

from magnify.registry import components


class Stitcher:
    def __init__(self, overlap: int = 102):
        self.overlap = overlap

    def __call__(self, assay: xr.Dataset) -> xr.Dataset:
        tiles = assay.tile[..., : -self.overlap, : -self.overlap]
        # Move the time and channel axes last so we can focus on joining images.
        tiles = tiles.transpose("tile_row", "tile_col", "tile_y", "tile_x", "channel", "time")
        tiles = xr.concat(tiles, dim="tile_y")
        images = xr.concat(tiles, dim="tile_x")
        # Change the x and y dimension names to be about the images.
        images = images.rename(tile_y="im_y", tile_x="im_x")
        # Move the time and channel axes back to the front.
        images = images.transpose("channel", "time", "im_y", "im_x")
        # Add the stitched images to the dataset.
        assay["image"] = images
        return assay

    @components.register("stitch")
    def make(overlap: int = 102):
        return Stitcher(overlap=overlap)



from __future__ import annotations

import math
import logging

from numpy.typing import ArrayLike
import cv2 as cv
import dask.array as da
import numba
import numpy as np
import scipy
import sklearn.mixture
import tqdm
import xarray as xr

from magnify import utils
import magnify.registry as registry

logger = logging.getLogger(__name__)


class ButtonFinder:
    def __init__(
        self,
        row_dist: float = 375 / 3.22,
        col_dist: float = 655 / 3.22,
        min_button_radius: int = 4,
        max_button_radius: int = 15,
        min_contrast: int | None = None,
        min_roundness: float = 0.75,
        cluster_penalty: float = 50,
        roi_length: int = 61,
        progress_bar: bool = False,
        search_timestep: list[int] | None = None,
        search_channel: str | list[str] | None = None,
    ):
        self.row_dist = row_dist
        self.col_dist = col_dist
        self.min_button_radius = min_button_radius
        self.max_button_radius = max_button_radius
        self.min_roundness = min_roundness
        self.cluster_penalty = cluster_penalty
        self.roi_length = roi_length
        self.progress_bar = progress_bar
        self.search_timesteps = sorted(utils.to_list(search_timestep)) if search_timestep else [0]
        self.search_channels = utils.to_list(search_channel)
        self.min_contrast = min_contrast

    def __call__(self, assay: xr.Dataset) -> xr.Dataset:
        if not self.search_channels:
            self.search_channels = assay.channel

        num_rows, num_cols = assay.tag.shape

        # Store each channel and timesteps for each marker in one chunk and set marker row/col
        # sizes so each chunk ends up being at least 50MB. We will rechunk later.
        chunk_bytes = 5e7
        # Don't take into account dtype size since fg/bg bool arrays should also be 50MB.
        roi_bytes = self.roi_length**2
        # Prioritize larger row chunks since we're more likely to want whole columns than rows.
        row_chunk_size = min(math.ceil(chunk_bytes / roi_bytes), num_rows)
        col_chunk_size = math.ceil(chunk_bytes / (roi_bytes * row_chunk_size))
        # Create the array of subimage regions.
        roi = da.empty(
            (
                num_rows,
                num_cols,
                assay.dims["channel"],
                assay.dims["time"],
                self.roi_length,
                self.roi_length,
            ),
            dtype=assay.image.dtype,
            chunks=(
                row_chunk_size,
                col_chunk_size,
                1,
                1,
                self.roi_length,
                self.roi_length,
            ),
        )
        assay["roi"] = (("mark_row", "mark_col", "channel", "time", "roi_y", "roi_x"), roi)
        assay = assay.assign_coords(
            fg=(
                ("mark_row", "mark_col", "channel", "time", "roi_y", "roi_x"),
                da.empty_like(
                    roi,
                    dtype=bool,
                ),
            ),
            bg=(
                ("mark_row", "mark_col", "channel", "time", "roi_y", "roi_x"),
                da.empty_like(
                    roi,
                    dtype=bool,
                ),
            ),
            x=(
                ("mark_row", "mark_col", "time"),
                np.empty((num_rows, num_cols, assay.dims["time"])),
            ),
            y=(
                ("mark_row", "mark_col", "time"),
                np.empty((num_rows, num_cols, assay.dims["time"])),
            ),
        )

        # Run the button finding algorithm for each timestep specified in search_timesteps.
        for t in tqdm.tqdm(self.search_timesteps, disable=not self.progress_bar):
            # Preload all images for this timestep so we only read from disk once.
            images = assay.image.isel(time=t).compute()
            # Find button centers.
            assay.x[..., t], assay.y[..., t] = self.find_centers(
                images.sel(channel=self.search_channels), assay
            )

            # Compute the roi, foreground and background masks for all buttons.
            (
                assay.roi[:, :, :, t],
                assay.fg[:, :, :, t],
                assay.bg[:, :, :, t],
                assay.x[..., t],
                assay.y[..., t],
                assay.valid[..., t],
            ) = self.find_rois(images, t, assay)
            # Eagerly compute the roi values so the dask task graph doesn't get too large.
            assay["roi"] = assay.roi.persist()
            assay["fg"] = assay.fg.persist()
            assay["bg"] = assay.bg.persist()

        # Now fill in the remaining timesteps where we aren't searching.
        for t, time in enumerate(tqdm.tqdm(assay.time, disable=not self.progress_bar)):
            if t in self.search_timesteps:
                # Skip this timestep since we've already processed it.
                continue

            if t < self.search_timesteps[0]:
                # Backfill timesteps that come before the first searched timestep.
                copy_t = self.search_timesteps[0]
            else:
                # Re-use the button locations of the timestep just before this one since it's either
                # a searched timestep or a timestep we just copied locations into.
                copy_t = t - 1

            # Preload all images for this timestep so we only read from disk once and
            # convert all relevant data to numpy arrays since iterating through xarrays is slow.
            images = assay.image.sel(time=time).to_numpy()
            x = assay.x[..., copy_t].to_numpy()
            y = assay.y[..., copy_t].to_numpy()
            roi = np.empty_like(assay.roi[:, :, :, t])
            # Update the roi since the location is copied but the roi images are timestep specific.
            for i in range(num_rows):
                for j in range(num_cols):
                    top, bottom, left, right = utils.bounding_box(
                        round(x[i, j]),
                        round(y[i, j]),
                        roi.shape[-1],
                        assay.sizes["im_x"],
                        assay.sizes["im_y"],
                    )
                    roi[i, j] = images[..., top:bottom, left:right]

            assay.roi[:, :, :, t] = roi
            assay.fg[:, :, :, t] = assay.fg[:, :, :, copy_t]
            assay.bg[:, :, :, t] = assay.bg[:, :, :, copy_t]
            assay.x[..., t] = x
            assay.y[..., t] = y
            assay.valid[..., t] = assay.valid[..., copy_t]
            # Eagerly compute the roi values so the dask task graph doesn't get too large.
            assay["roi"] = assay.roi.persist()
            assay["fg"] = assay.fg.persist()
            assay["bg"] = assay.bg.persist()
        assay = assay.stack(mark=("mark_row", "mark_col"), create_index=True).transpose("mark", ...)
        # Rechunk the array to chunk along markers since users will usually want to slice along that dimension.
        mark_chunk_size = min(
            math.ceil(chunk_bytes / (roi_bytes * assay.dims["time"] * assay.dims["channel"])),
            num_rows,
        )
        chunk_sizes = [
            mark_chunk_size,
            assay.dims["channel"],
            assay.dims["time"],
            assay.dims["roi_y"],
            assay.dims["roi_x"],
        ]
        # Eagerly compute the rechunking to prevent delays.
        assay["roi"] = assay.roi.chunk(chunk_sizes).persist()
        assay["fg"] = assay.fg.chunk(chunk_sizes).persist()
        assay["bg"] = assay.bg.chunk(chunk_sizes).persist()

        return assay

    def find_centers(self, images: xr.DataArray, assay: xr.Dataset):
        points = np.empty((0, 2))
        min_button_dist = round(min(self.row_dist, self.col_dist) / 2)
        if min_button_dist % 2 == 0:
            # Certain opencv functions require an odd blocksize.
            min_button_dist -= 1
        for image in images:
            image_max = image.max().item()
            image = utils.to_uint8(image.to_numpy())
            # Step 1: Find an imperfect button mask by thresholding.
            blur = cv.medianBlur(image, 2 * min_button_dist + 1)
            if self.min_contrast is None:
                # Guess a reasonable threshold for buttons.
                num_points = (assay.tag != "").sum()
                q = np.pi * self.max_button_radius**2 * num_points / image.size
                thresh = np.quantile(np.maximum(image.astype(np.int16) - blur, 0), 1 - 2 * q)
            else:
                thresh = 255 * self.min_contrast / image_max
            mask = utils.to_uint8(image > blur + thresh)

            # Step 2: Find connected components and filter out points.
            _, _, stats, new_points = cv.connectedComponentsWithStats(mask, connectivity=4)

            # Ignore the background point.
            new_points = new_points[1:]
            stats = stats[1:]

            # Exclude large and small blobs.
            new_points = new_points[
                (stats[:, cv.CC_STAT_HEIGHT] <= 2 * self.max_button_radius)
                & (stats[:, cv.CC_STAT_WIDTH] <= 2 * self.max_button_radius)
                & (stats[:, cv.CC_STAT_AREA] <= np.pi * self.max_button_radius**2)
                & (stats[:, cv.CC_STAT_HEIGHT] >= 2 * self.min_button_radius)
                & (stats[:, cv.CC_STAT_WIDTH] >= 2 * self.min_button_radius)
                & (stats[:, cv.CC_STAT_AREA] >= np.pi * self.min_button_radius**2)
            ]
            # Remove points too close to other points in this channel.
            dist_matrix = np.linalg.norm(new_points[np.newaxis] - new_points[:, np.newaxis], axis=2)
            dist_matrix[np.diag_indices(len(dist_matrix))] = np.inf
            new_points = new_points[np.min(dist_matrix, axis=1) > min_button_dist]
            if len(points) > 0:
                # Remove points too close to other points in previous channels.
                dist_matrix = np.linalg.norm(points[np.newaxis] - new_points[:, np.newaxis], axis=2)
                new_points = new_points[np.min(dist_matrix, axis=1) > min_button_dist]

            # Add the new points to the list of seen points.
            points = np.concatenate([points, new_points])

        # Split the points into x and y components.
        x = points[:, 0]
        y = points[:, 1]

        # Step 3: Cluster the points into distinct rows and columns.
        points_per_row = (assay.tag != "").sum(dim="mark_col").to_numpy()
        points_per_col = (assay.tag != "").sum(dim="mark_row").to_numpy()
        num_rows, num_cols = assay.sizes["mark_row"], assay.sizes["mark_col"]
        row_labels = cluster_1d(
            y,
            total_length=image.shape[0],
            num_clusters=num_rows,
            cluster_length=self.row_dist,
            ideal_num_points=points_per_row,
            penalty=self.cluster_penalty,
        )
        col_labels = cluster_1d(
            x,
            total_length=image.shape[1],
            num_clusters=num_cols,
            cluster_length=self.col_dist,
            ideal_num_points=points_per_col,
            penalty=self.cluster_penalty,
        )

        # Exclude boundary points that didn't fall into clusters.
        in_cluster = (row_labels >= 0) & (col_labels >= 0)
        x, y = x[in_cluster], y[in_cluster]
        col_labels, row_labels = col_labels[in_cluster], row_labels[in_cluster]

        # Step 4: Draw lines through each cluster.
        row_slope, row_intercepts = regress_clusters(
            x, y, labels=row_labels, num_clusters=num_rows, ideal_num_points=points_per_row
        )
        # We treat column indices as y and row indices as x to avoid near-infinite slopes.
        col_slope, col_intercepts = regress_clusters(
            y,
            x,
            labels=col_labels,
            num_clusters=num_cols,
            ideal_num_points=points_per_col,
        )

        # Step 5: Set button locations as the intersection of each line pair.
        mark_y = (row_slope * col_intercepts[np.newaxis] + row_intercepts[:, np.newaxis]) / (
            1 - row_slope * col_slope
        )
        mark_x = mark_y * col_slope + col_intercepts[np.newaxis]

        return mark_x, mark_y

    def find_rois(self, images: xr.DataArray, t: int, assay: xr.Dataset):
        # Convert all relevant quantities to numpy arrays since xarrays are very slow
        # when iterated over.
        images = images.to_numpy()
        num_rows, num_cols = assay.roi.shape[:2]
        tag = assay.tag.to_numpy()
        x = assay.x[:, :, t].to_numpy()
        y = assay.y[:, :, t].to_numpy()
        valid = assay.valid[:, :, t].to_numpy()
        roi = np.empty_like(assay.roi.isel(time=t))
        fg = np.empty_like(roi, dtype=bool)
        bg = np.empty_like(fg)
        search_channel_idxs = [
            list(assay.channel.to_numpy()).index(c) for c in self.search_channels
        ]
        offsets = np.zeros((num_rows, num_cols, 2), dtype=int)

        for i in range(num_rows):
            for j in range(num_cols):
                # Initialize the roi image.
                top, bottom, left, right = utils.bounding_box(
                    round(x[i, j]),
                    round(y[i, j]),
                    roi.shape[-1],
                    assay.sizes["im_x"],
                    assay.sizes["im_y"],
                )
                roi[i, j] = images[..., top:bottom, left:right]

                best_subimage = None
                best_contour = None
                best_roundness = 0
                # Refine our button estimate unless we have a blank chamber.
                if tag[i, j] != "":
                    for channel in search_channel_idxs:
                        subimage = roi[i, j, channel]
                        subimage = utils.to_uint8(np.clip(subimage, np.median(subimage), None))
                        subimage -= subimage.min()
                        if self.min_contrast is None:
                            _, mask = cv.threshold(
                                subimage, thresh=0, maxval=1, type=cv.THRESH_BINARY + cv.THRESH_OTSU
                            )
                        else:
                            thresh = 255 * self.min_contrast / roi[i, j, channel].max()
                            mask = utils.to_uint8(subimage > thresh)

                        contours, hierarchy = cv.findContours(
                            mask, cv.RETR_CCOMP, cv.CHAIN_APPROX_SIMPLE
                        )
                        hierarchy = hierarchy[0]
                        for c, h in zip(contours, hierarchy):
                            # Don't consider holes.
                            if h[3] == 0:
                                continue
                            perimeter = cv.arcLength(c, True)
                            area = cv.contourArea(c)
                            if h[2] != -1:
                                area -= cv.contourArea(contours[h[2]])
                                # Save the contour of the hole too.
                                c = [c, contours[h[2]]]
                            else:
                                # Save contour as element of a list for consistency.
                                c = [c]
                            # Don't consider contours that are the wrong size.
                            if (
                                perimeter <= 2 * np.pi * self.min_button_radius
                                or perimeter >= 2 * np.pi * self.max_button_radius
                                or area <= np.pi * self.min_button_radius**2
                                or area >= np.pi * self.max_button_radius**2
                            ):
                                continue
                            # Save this contour if it's the roundest one.
                            roundness = 4 * np.pi * area / perimeter**2
                            if roundness > best_roundness and roundness > self.min_roundness:
                                best_roundness = roundness
                                best_contour = c
                                best_subimage = subimage

                # Update our estimate of the button position if we found some circles.
                if best_contour is not None:
                    y[i, j], x[i, j] = utils.contour_center(best_contour[0])
                    # Refine the background by removing bright areas.
                    # Change coordinates from roi to image coordinates.
                    x[i, j] += left
                    y[i, j] += top
                    # Move the roi bounding box to center the new x, y values.
                    old_top = top
                    old_left = left
                    top, bottom, left, right = utils.bounding_box(
                        round(x[i, j]),
                        round(y[i, j]),
                        self.roi_length,
                        assay.sizes["im_x"],
                        assay.sizes["im_y"],
                    )
                    roi[i, j] = images[..., top:bottom, left:right]
                    # Move contour coordinates to new centers.
                    for c in best_contour:
                        c[:, :, 0] += old_left - left
                        c[:, :, 1] += old_top - top

                x_rel = round(x[i, j]) - left
                y_rel = round(y[i, j]) - top

                # Set the background to be the annulus around a circle of fixed radius.
                bg_mask = utils.annulus(
                    self.roi_length,
                    row=y_rel,
                    col=x_rel,
                    outer_radius=2 * self.max_button_radius,
                    inner_radius=self.max_button_radius,
                    value=1,
                )

                if best_contour is not None:
                    fg_mask = np.zeros(fg.shape[-2:], dtype=np.uint8)
                    cv.drawContours(fg_mask, best_contour, -1, 1, cv.FILLED)
                    # Refine the background by removing bright areas.
                    _, bright_mask = cv.threshold(
                        best_subimage, thresh=0, maxval=1, type=cv.THRESH_BINARY + cv.THRESH_OTSU
                    )
                    dim_mask = 1 - cv.dilate(
                        bright_mask, np.ones((self.max_button_radius, self.max_button_radius))
                    )
                    if np.any(bg_mask * dim_mask):
                        bg_mask *= dim_mask
                    valid[i, j] = True
                else:
                    # If we didn't find a suitable foreground just set it to be a large circle.
                    fg_mask = utils.circle(
                        self.roi_length,
                        row=y_rel,
                        col=x_rel,
                        radius=self.max_button_radius,
                        value=1,
                    )
                    valid[i, j] = False

                fg[i, j] = fg_mask
                bg[i, j] = bg_mask

        return roi, fg, bg, x, y, valid

    @registry.components.register("find_buttons")
    def make(
        row_dist: float = 375 / 3.22,
        col_dist: float = 655 / 3.22,
        min_button_radius: int = 4,
        max_button_radius: int = 15,
        min_contrast: int | None = None,
        min_roundness: float = 0.75,
        cluster_penalty: float = 50,
        roi_length: int = 61,
        progress_bar: bool = False,
        search_timestep: list[int] | None = None,
        search_channel: str | list[str] | None = None,
    ):
        return ButtonFinder(
            row_dist=row_dist,
            col_dist=col_dist,
            min_button_radius=min_button_radius,
            max_button_radius=max_button_radius,
            min_contrast=min_contrast,
            min_roundness=min_roundness,
            cluster_penalty=cluster_penalty,
            roi_length=roi_length,
            progress_bar=progress_bar,
            search_timestep=search_timestep,
            search_channel=search_channel,
        )


class BeadFinder:
    def __init__(
        self,
        min_bead_radius: int = 5,
        max_bead_radius: int = 25,
        roi_length: int = 61,
        search_channel: str | list[str] | None = None,
    ):
        self.min_bead_radius = min_bead_radius
        self.max_bead_radius = max_bead_radius
        self.roi_length = roi_length
        self.search_channels = utils.to_list(search_channel)

    def __call__(self, assay: xr.Dataset) -> xr.Dataset:
        if not self.search_channels:
            self.search_channels = assay.channel

        centers = np.empty((0, 2))
        labels = np.zeros((assay.sizes["im_y"], assay.sizes["im_x"]), dtype=int)
        for t in assay.time:
            for search_channel in self.search_channels:
                image = utils.to_uint8(assay.image.sel(channel=search_channel, time=t).to_numpy())
                # Step 1: Denoise the image for more accurate edge finding.
                image = cv.GaussianBlur(image, (5, 5), 0)

                # Step 2: Find edges from image gradients.
                dx = cv.Scharr(image, ddepth=cv.CV_16S, dx=1, dy=0)
                dy = cv.Scharr(image, ddepth=cv.CV_16S, dx=0, dy=1)
                grad = np.abs(dx) + np.abs(dy)
                # Assuming at least 1% of the image contains beads the local maxima will usually
                # be an edge so use that to set a reasonable upper threshold for edge finding.
                grad_max = cv.dilate(
                    grad, np.ones((10 * self.max_bead_radius, 10 * self.max_bead_radius))
                )
                # The median gradient should be background so use it as the lower threshold.
                edges = 255 - cv.Canny(
                    dx=dx, dy=dy, threshold1=np.median(grad), threshold2=np.median(grad_max) / 3
                )

                # Step 3: Treat edges as boundaries to find connected components that are either
                # all background or foreground.
                edge_width = 2 * self.min_bead_radius
                kernel = cv.getStructuringElement(
                    cv.MORPH_ELLIPSE, (edge_width + 1, edge_width + 1)
                )
                # Thicken edges to make sure they will be closed.
                edges = cv.erode(edges, kernel)
                _, l, stats, c = cv.connectedComponentsWithStats(
                    edges, connectivity=8, ltype=cv.CV_16U
                )

                # Step 4: Filter out regions that don't meet size criteria or aren't bright enough.
                min_radius = self.min_bead_radius - edge_width
                # Take into account the fact that beads are smaller because we grew edges.
                is_fg = (
                    (stats[:, cv.CC_STAT_WIDTH] >= min_radius)
                    & (stats[:, cv.CC_STAT_HEIGHT] >= min_radius)
                    & (stats[:, cv.CC_STAT_AREA] >= 2 * np.pi * min_radius)
                    & (stats[:, cv.CC_STAT_WIDTH] <= 2 * self.max_bead_radius + 200)
                    & (stats[:, cv.CC_STAT_HEIGHT] <= 2 * self.max_bead_radius)
                    & (stats[:, cv.CC_STAT_AREA] <= np.pi * self.max_bead_radius**2)
                )

                # Estimate the median background and foreground intensity.
                fg_vals = []
                bg_vals = []
                for i in range(1, l.max() + 1):
                    if is_fg[i]:
                        fg_vals.append(np.median(image[l == i]))
                    else:
                        bg_vals.append(image[l == i])
                fg_vals = np.array(fg_vals)
                bg_vals = np.concatenate(bg_vals)
                fg_median = np.median(fg_vals)
                bg_median = np.median(bg_vals)

                # Exclude foreground components whose brightness is similar to background if the
                # distribution of foregrounds looks bimodal.
                vals = np.concatenate([fg_vals, bg_vals[: len(fg_vals)]])[:, np.newaxis]
                fg_vals = fg_vals[:, np.newaxis]
                is_fg[is_fg] = (
                    sklearn.mixture.GaussianMixture(
                        n_components=2, means_init=([[bg_median], [fg_median]])
                    )
                    .fit(vals)
                    .predict(fg_vals)
                    == 1
                )

                # Reindex labels to be contiguous and set all non-foregrounds to 0.
                label_idx = 1
                for i in range(1, l.max() + 1):
                    if is_fg[i]:
                        l[l == i] = label_idx
                        label_idx += 1
                    else:
                        # TODO: We should handle edges differently since they're not backgrounds.
                        l[l == i] = 0

                # Redilate all foreground components to account for the erosion we did earlier.
                l = cv.dilate(l, kernel)

                # Step 5: Add new beads to previous channels' results.
                c = c[is_fg]
                if len(centers) > 0:
                    # Exclude beads that we've already seen.
                    duplicates = np.array(
                        [
                            len(neighbors) > 0
                            for neighbors in scipy.spatial.KDTree(centers).query_ball_point(
                                c, 2 * self.min_bead_radius
                            )
                        ]
                    )
                    c = c[~duplicates]
                    l[np.isin(l, np.where(duplicates)[0] + 1)] = 0
                    l = np.unique(l, return_inverse=True)[1].reshape(l.shape)
                centers = np.concatenate([centers, c])
                nnz = l != 0
                labels[nnz] = l[nnz] + labels.max()

        num_beads = len(centers)
        # Store each channel and timesteps for each marker in one chunk and set marker row/col
        # sizes so each chunk ends up being at least 50MB. We will rechunk later.
        chunk_bytes = 5e7
        # Don't take into account dtype size since fg/bg bool arrays should also be 50MB.
        roi_bytes = self.roi_length**2
        # Create the array of subimage regions.
        roi = da.empty(
            (
                num_beads,
                assay.dims["channel"],
                assay.dims["time"],
                self.roi_length,
                self.roi_length,
            ),
            dtype=assay.image.dtype,
            chunks=(
                min(
                    math.ceil(
                        chunk_bytes / (roi_bytes * assay.dims["channel"] * assay.dims["time"])
                    ),
                    num_beads,
                ),
                assay.dims["channel"],
                assay.dims["time"],
                self.roi_length,
                self.roi_length,
            ),
        )

        assay["roi"] = (("mark", "channel", "time", "roi_y", "roi_x"), roi)
        assay = assay.assign_coords(
            fg=(
                ("mark", "channel", "time", "roi_y", "roi_x"),
                da.empty_like(roi, dtype=bool),
            ),
            bg=(
                ("mark", "channel", "time", "roi_y", "roi_x"),
                da.empty_like(roi, dtype=bool),
            ),
            x=(
                ("mark", "time"),
                np.repeat(centers[:, np.newaxis, 0], assay.dims["time"], axis=1),
            ),
            y=(
                ("mark", "time"),
                np.repeat(centers[:, np.newaxis, 1], assay.dims["time"], axis=1),
            ),
            valid=(
                ("mark", "time"),
                np.ones((assay.sizes["mark"], assay.sizes["time"]), dtype=bool),
            ),
        )

        # Compute the foreground and background masks for all buttons.
        # TODO: Don't assume beads don't move across timesteps.
        # Iterate over numpy arrays since indexing over xarrays is slow.
        x = assay.x.sel(time=0).to_numpy()
        y = assay.y.sel(time=0).to_numpy()
        fg = np.empty((num_beads,) + assay.fg.shape[2:], dtype=bool)
        bg = np.empty_like(fg)
        for i in range(num_beads):
            # Set the subimage region for this bead.
            top, bottom, left, right = utils.bounding_box(
                round(x[i]),
                round(y[i]),
                self.roi_length,
                assay.sizes["im_x"],
                assay.sizes["im_y"],
            )
            # Set the foreground of the bead.
            fg[i] = labels[top:bottom, left:right] == i + 1
            # Set the background to be the region assigned to no beads.
            bg[i] = labels[top:bottom, left:right] == 0
        assay.fg[:] = fg[:, np.newaxis]
        assay.bg[:] = bg[:, np.newaxis]

        for i, channel in enumerate(assay.channel):
            image = assay.image.sel(channel=channel).to_numpy()
            roi = np.empty((num_beads,) + assay.roi.shape[2:], dtype=assay.roi.dtype)
            for j in range(num_beads):
                # Set the subimage region for this bead.
                top, bottom, left, right = utils.bounding_box(
                    round(x[j]),
                    round(y[j]),
                    self.roi_length,
                    assay.sizes["im_x"],
                    assay.sizes["im_y"],
                )
                roi[j] = image[..., top:bottom, left:right]
            assay.roi[:, i] = roi

        assay["roi"] = assay.roi.persist()
        assay["fg"] = assay.fg.persist()
        assay["bg"] = assay.bg.persist()
        assay = assay.assign_coords()
        return assay

    @registry.components.register("find_beads")
    def make(
        min_bead_radius: int = 5,
        max_bead_radius: int = 25,
        roi_length: int = 61,
        search_channel: str | list[str] | None = None,
    ):
        return BeadFinder(
            min_bead_radius=min_bead_radius,
            max_bead_radius=max_bead_radius,
            roi_length=roi_length,
            search_channel=search_channel,
        )


def cluster_1d(
    points: np.ndarray,
    total_length: int,
    num_clusters: int,
    cluster_length: float,
    ideal_num_points: np.ndarray,
    penalty: float,
) -> np.ndarray:
    # Find the best clustering using the accumulate ragged array trick.
    # See: https://vladfeinberg.com/2021/01/07/vectorizing-ragged-arrays.html
    permutation = np.argsort(points)
    points = points[permutation]

    min_cost = np.inf
    best_spans = None
    for offset in range(total_length - round(num_clusters * cluster_length)):
        # Compute the boundaries and center of each cluster.
        boundaries = np.arange(num_clusters + 1) * cluster_length + offset
        centers = (boundaries[1:] + boundaries[:-1]) / 2
        # Find start/end indexes and number of points in each cluster.
        spans = np.searchsorted(points, boundaries)
        num_points = spans[1:] - spans[:-1]
        # Compute the squared distance of each point that falls in a cluster to its cluster center.
        dists = (points[spans[0] : spans[-1]] - np.repeat(centers, num_points)) ** 2
        # Penalize clusters for high variance.
        cost = np.insert(np.cumsum(dists), 0, 0)
        cost = np.diff(cost[spans - spans[0]])
        cost[num_points > 0] /= num_points[num_points > 0]
        # Set empty clusters to the maximum variance.
        cost[num_points == 0] = np.max(cost)
        # Fewer points in a cluster means noisier variances so adjust variance contributions.
        cost *= np.sqrt(ideal_num_points)
        # Penalize clusters for having too few or too many points.
        cost = cost + penalty * (ideal_num_points - num_points) ** 2
        if cost.sum() < min_cost:
            min_cost = cost.sum()
            best_spans = spans

    # Label each point with its cluster, label points outside clusters as -1.
    labels = -np.ones_like(points, dtype=int)
    labels[best_spans[0] : best_spans[-1]] = np.repeat(
        np.arange(num_clusters), best_spans[1:] - best_spans[:-1]
    )

    # Return the labels based on the original order of the points.
    return labels[np.argsort(permutation)]


def regress_clusters(
    x: np.ndarray,
    y: np.ndarray,
    labels: np.ndarray,
    num_clusters: int,
    ideal_num_points: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    # Find the best line per-cluster.
    slopes = np.full(num_clusters, np.nan)
    intercepts = np.full(num_clusters, np.nan)
    cluster_points = [(x[labels == i], y[labels == i]) for i in range(num_clusters)]
    for i, (x, y) in enumerate(cluster_points):
        # Only regress on multi-point clusters.
        if len(x) > 1:
            slopes[i], intercepts[i], _, _, _ = scipy.stats.linregress(x, y)
        elif (i == 0 or i == num_clusters - 1) and ideal_num_points[i] >= 2:
            logger.warning(
                "Boundary cluster has fewer than 2 points."
                "The chip is unlikely to be segmented correctly."
            )

    # Recompute the intercepts using the median slope.
    slope = np.nanmedian(slopes)
    for i, (x, y) in enumerate(cluster_points):
        if len(x) > 0:
            intercepts[i] = np.median(y - slope * x)

    # Globally estimate where intercepts are, assuming they're evenly spaced.
    not_nan = ~np.isnan(intercepts)
    label_idxs = np.arange(num_clusters)
    intercept_m, intercept_b, _, _, _ = scipy.stats.linregress(
        label_idxs[not_nan], intercepts[not_nan]
    )
    # Re-estimate intercepts using a weighted mean of global and local estimates.
    # This reduces outlier effects while still allowing uneven intercepts from image stitching.
    for i, (x, y) in enumerate(cluster_points):
        if ideal_num_points[i] != 0 and not_nan[i]:
            weight = min(len(x), ideal_num_points[i]) / ideal_num_points[i]
            intercepts[i] = weight * intercepts[i] + (1 - weight) * (intercept_m * i + intercept_b)
        else:
            # Just use our global estimate when we have an empty cluster.
            intercepts[i] = intercept_m * i + intercept_b

    return slope, intercepts


@numba.jit(nopython=True)
def get_label_centers(labels):
    num_labels = labels.max() + 1
    centers = np.zeros((num_labels, 2))
    num_points = np.zeros(num_labels)
    for i in range(labels.shape[0]):
        for j in range(labels.shape[1]):
            k = labels[i, j]
            centers[k, 0] += j
            centers[k, 1] += i
            num_points[k] += 1
    centers /= np.expand_dims(num_points, axis=1)
    return centers[1:]



from __future__ import annotations
import re

import numpy as np
import pandas as pd
import scipy

import magnify.registry as registry


@registry.component("identify_buttons")
def identify_buttons(assay, pinlist, blank=None):
    if blank is None:
        blank = ["", "blank", "BLANK"]

    df = pd.read_csv(pinlist)
    df["Indices"] = df["Indices"].apply(
        lambda s: [int(x) for x in re.sub(r"[\(\)]", "", s).split(",")]
    )
    # Replace blanks with the empty string.
    df["MutantID"] = df["MutantID"].replace(blank, "")
    # Zero-index the indices.
    cols, rows = np.array(df["Indices"].to_list()).T - 1
    names = df["MutantID"].to_numpy(dtype=str, na_value="")
    names_array = np.empty((max(rows) + 1, max(cols) + 1), dtype=names.dtype)
    names_array[rows, cols] = names
    assay = assay.assign_coords(
        tag=(("mark_row", "mark_col"), names_array),
        valid=(
            ("mark_row", "mark_col", "time"),
            np.ones((names_array.shape[0], names_array.shape[1], assay.sizes["time"]), dtype=bool),
        ),
    )
    return assay


@registry.component("identify_mrbles")
def indentify_mrbles(assay, spectra, codes, reference="eu"):
    # Read in the dataframe of lanthanide spectra and make sure the reference lanthanide is first.
    spectra_df = pd.read_csv(spectra)
    ref_idx = spectra_df[spectra_df["name"] == reference].index[0]
    spectra_df = spectra_df.reindex([ref_idx] + [i for i in range(len(spectra_df)) if i != ref_idx])
    lns = spectra_df["name"].to_list()
    num_lns = len(lns)

    # Read in the dataframe of codes.
    codes_df = pd.read_csv(codes)
    tag_names = codes_df["name"].to_numpy()
    num_codes = len(tag_names)
    # Make sure spectra and codes have consistent lanthanides.
    code_lns = set(codes_df.columns)
    code_lns.remove("name")
    if code_lns != set(lns):
        raise ValueError(f"Lanthanide names in {codes} do not match lanthanide names in {spectra}.")

    # Step 1: Estimate the lanthanide volumes in each bead by solving the linear equation SV = I
    # where S are the reference spectra and I are the intensities of each bead.
    channels = [c for c in assay.channel.values if c in spectra_df.columns]
    sp = spectra_df[channels].to_numpy()
    sel = assay.roi.isel(time=0).sel(channel=channels)
    intensities = sel.where(sel.fg).mean(dim=["roi_x", "roi_y"]).to_numpy()
    volumes = np.linalg.lstsq(sp.T, intensities.T, rcond=None)[0].T
    # We also want the lanthanide ratios with respect to the reference lanthanide.
    ratios = volumes / volumes[:, 0:1]
    # Save the resulting lanthanide volumes and ratios to the assay.
    assay = assay.assign_coords(ln=("ln", lns))
    assay["ln_vol"] = (("mark", "ln"), volumes)
    assay["ln_ratio"] = (("mark", "ln"), ratios)

    # Step 2: Agressively remove outliers to make future processing easier.
    X = ratios[:, 1:]
    # Find the distance to a point that should still be in the same cluster assuming cluster
    # sizes differ by a factor of at most 20 from the mean cluster size.
    n_neighbor = round(len(X) / (20 * num_codes)) + 2
    dist = (
        scipy.spatial.KDTree(X, leafsize=n_neighbor)
        .query(X, k=[n_neighbor], workers=-1)[0]
        .flatten()
    )
    # We care more about excluding all outliers so exclude 5% of points.
    X_r = X[dist <= np.percentile(dist, 95)]

    # Step 3: Find an affine transformation of the code's lanthanide ratios to get a clustering
    # that minimizes the distance between each bead and its closest code.
    code_ratios = codes_df[lns[1:]].to_numpy()

    # We will try to find a good affine transformation by minimizing a function that approximates
    # a per-cluster distance function.
    def loss(theta):
        A = theta[: num_lns - 1]
        p = theta[num_lns - 1 :]
        eps = 1e-8
        dist = np.linalg.norm((A * code_ratios + p)[np.newaxis] - X_r[:, np.newaxis], axis=-1)
        # Logsumexp is a smooth approximation to the max function when eps is small.
        return -eps * np.sum(scipy.special.logsumexp(-dist / eps, axis=-1)) / len(X_r)

    # Minimize the loss only considering affine transforms close to an estimated scaling factor.
    scales = (X_r.max(axis=0) - X_r.min(axis=0)) / (
        code_ratios.max(axis=0) - code_ratios.min(axis=0)
    )
    theta = scipy.optimize.minimize(
        loss,
        x0=np.concatenate((scales, X_r.min(axis=0) - scales * code_ratios.min(axis=0))),
        bounds=np.concatenate(
            (
                (
                    np.column_stack((0.75 * scales, 1.25 * scales)),
                    np.column_stack(
                        (
                            (X_r.min(axis=0) - 1.25 * scales * code_ratios.min(axis=0)),
                            X_r.max(axis=0) - 0.75 * scales * code_ratios.max(axis=0),
                        )
                    ),
                )
            )
        ),
    ).x
    A = theta[: num_lns - 1]
    p = theta[num_lns - 1 :]

    # Cluster points to the closest code.
    tag_idxs = np.argmin(
        np.linalg.norm(X_r[:, np.newaxis] - (A * code_ratios + p)[np.newaxis], axis=-1), axis=1
    )

    # Change the number of codes to be the number of clusters we found.
    nonzero_tags, counts = np.unique(tag_idxs, return_counts=True)
    nonzero_tags = nonzero_tags[counts >= 5]
    num_codes = len(nonzero_tags)

    # Step 4: Perform a better clustering using a Gaussian mixture model initialized with the
    # clustering from step 3. We also add a uniform distribution to the mixture which allows us to
    # exclude outliers less agressively.
    means = np.zeros((num_codes, num_lns - 1))
    covs = np.zeros((num_codes, num_lns - 1, num_lns - 1))
    proportions = np.zeros(num_codes + 1)
    # Initialize the Gaussian components.
    for i, tag_idx in enumerate(nonzero_tags):
        proportions[i] = np.sum(tag_idxs == tag_idx)
        means[i] = np.median(X_r[tag_idxs == tag_idx], axis=0)
        covs[i] = np.cov(X_r[tag_idxs == tag_idx], rowvar=False)
    # Initialize the uniform component.
    proportions[-1] = len(X) - len(X_r)
    proportions /= proportions.sum()
    log_cond_probs = np.empty((len(X), num_codes + 1))
    log_cond_probs[:, -1] = -np.log(X.max(axis=0) - X.min(axis=0)).sum()
    probs = None

    # Run the Expectation-Maximization algorithm.
    for i in range(30):
        # E-step: Compute the probability of each point belonging to each component.
        diff = X[:, np.newaxis, :] - means[np.newaxis, :, :]
        # Work in log space most of the time to avoid numerical issues.
        try:
            log_cond_probs[:, :-1] = (
                -X.shape[1] * np.log(2 * np.pi) / 2
                - 0.5 * np.log(np.linalg.det(covs))
                - 0.5 * np.einsum("...i,...ij,...j->...", diff, np.linalg.inv(covs), diff)
            )
        except np.linalg.LinAlgError:
            print("Warning: Code clustering did not converge.")
            break
        log_probs = np.log(proportions) + log_cond_probs
        log_probs -= scipy.special.logsumexp(log_probs, axis=1)[:, np.newaxis]
        probs = np.exp(log_probs)
        # M-step: Update the parameters of each component.
        means = (
            np.sum(probs[:, :-1, np.newaxis] * X[:, np.newaxis, :], axis=0)
            / np.sum(probs[:, :-1], axis=0)[:, np.newaxis]
        )
        diff = X[:, np.newaxis, :] - means[np.newaxis, :, :]
        covs = (
            np.sum(
                probs[:, :-1, np.newaxis, np.newaxis] * np.einsum("...i,...j->...ij", diff, diff),
                axis=0,
            )
            / np.sum(probs[:, :-1], axis=0)[:, np.newaxis, np.newaxis]
        )
        # Regularize the covariance matrix to avoid numerical issues.
        covs += np.eye(X.shape[1]) * 1e-6
        proportions = np.sum(probs, axis=0) / X.shape[0]

    # Assign each bead a code based on the clustering we just found.
    nonzero_tags = np.append(nonzero_tags, -1)
    tag_names = np.append(tag_names, "outlier")
    if probs is not None:
        tag_idxs = nonzero_tags[np.argmax(probs, axis=1)]
    else:
        tag_idxs = np.argmin(
            np.linalg.norm(X[:, np.newaxis] - (A * code_ratios + p)[np.newaxis], axis=-1), axis=1
        )
    assay = assay.assign_coords(
        tag=("mark", tag_names[tag_idxs]),
    )

    return assay



from __future__ import annotations
import functools
import inspect
from typing import Callable

from numpy.typing import ArrayLike
import catalogue
import confection

from magnify.pipeline import Pipeline

readers = catalogue.create("magnify", "readers")
components = catalogue.create("magnify", "components")


def component(name):
    def component_decorator(func):
        @functools.wraps(func)
        def component_factory(*args, **kwargs):
            return functools.partial(func, *args, **kwargs)

        # Make the factory signature identical to func's except for the first argument.
        signature = inspect.signature(func)
        signature = signature.replace(parameters=list(signature.parameters.values())[1:])
        component_factory.__signature__ = signature
        components.register(name)(component_factory)
        return func

    return component_decorator


def ps_chip(**kwargs):
    # Button centers are apart 375um vertically and 655um horizontally.
    # Assuming a 4x objective and 2x2 binning each pixel is 3.22um.
    defaults = confection.Config(dict(row_dist=375 / 3.22, col_dist=655 / 3.22))
    config = defaults.merge(confection.Config(kwargs))
    pipe = Pipeline("read", config=config)
    pipe.add_pipe("identify_buttons")
    pipe.add_pipe("horizontal_flip")
    pipe.add_pipe("stitch")
    pipe.add_pipe("horizontal_flip")
    pipe.add_pipe("find_buttons")
    pipe.add_pipe("filter_expression")
    pipe.add_pipe("filter_nonround")
    pipe.add_pipe("filter_leaky")
    pipe.add_pipe("drop")

    return pipe


def pc_chip(**kwargs):
    # Button centers are apart 412um vertically and 760um horizontally.
    # Assuming a 4x objective and 2x2 binning each pixel is 3.22um.
    defaults = confection.Config(dict(row_dist=406 / 3.22, col_dist=750 / 3.22))
    config = defaults.merge(confection.Config(kwargs))
    pipe = Pipeline("read", config=config)
    pipe.add_pipe("identify_buttons")
    pipe.add_pipe("horizontal_flip")
    pipe.add_pipe("stitch")
    pipe.add_pipe("horizontal_flip")
    pipe.add_pipe("find_buttons")
    pipe.add_pipe("filter_expression")
    pipe.add_pipe("filter_nonround")
    pipe.add_pipe("filter_leaky")
    pipe.add_pipe("drop")

    return pipe


def mrbles(**kwargs):
    pipe = Pipeline("read", config=kwargs)
    pipe.add_pipe("flatfield_correct")
    pipe.add_pipe("stitch")
    pipe.add_pipe("find_beads")
    pipe.add_pipe("circularize")
    pipe.add_pipe("identify_mrbles")
    pipe.add_pipe("drop")

    return pipe


def beads(**kwargs):
    pipe = Pipeline("read", config=kwargs)
    pipe.add_pipe("flatfield_correct")
    pipe.add_pipe("stitch")
    pipe.add_pipe("find_beads")
    pipe.add_pipe("circularize")
    pipe.add_pipe("drop")

    return pipe


def image(**kwargs):
    pipe = Pipeline("read", config=kwargs)
    pipe.add_pipe("stitch")
    pipe.add_pipe("drop")
    return pipe



from __future__ import annotations

import xarray as xr


def save(file, xp):
    # xarrays don't support saving multi-indices so unstack before saving.
    xp.unstack().to_netcdf(file)


def load(file):
    xp = xr.open_dataset(file)
    # If we have a chip dataset we want to restack the mark_col and mark_row indices
    # which we unstacked before saving.
    if "mark_row" in xp.dims and "mark_col" in xp.dims:
        xp = xp.stack(mark=("mark_row", "mark_col")).transpose("mark", ...)
    return xp



from __future__ import annotations

import numpy as np
import plotly.express as px
import scipy.stats
import xarray as xr

from magnify import utils
from magnify.plot.ndplot import ndplot


def relplot(
    assay: xr.Dataset,
    x="time",
    y="intensity",
    fit_method="linear",
    grid=None,
    hue=None,
    slider=None,
    **kwargs,
):
    if grid is None and slider is None:
        if y == "mark_intensity":
            slider = ["channel", "tag"]
        else:
            slider = ["channel", "tag"]
    fit_func = {"linear": linear_fit, "quadratic": quadratic_fit, "exp_linear": exp_linear_fit}[
        fit_method
    ]

    def relfunc(assay: xr.Dataset, **kwargs):
        assay = assay.where(assay.valid, drop=True)
        overlays = []
        return px.scatter(assay.to_dataframe().reset_index(), x=x, y=y, color=hue).data

    return ndplot(assay, relfunc, grid=grid, slider=slider, **kwargs)


def linear_fit(x, y):
    m, b, _, _, _ = scipy.stats.linregress(x, y)
    return m * x + b


def quadratic_fit(x, y):
    X = np.column_stack((x**2, x, np.ones(len(x))))
    theta = np.linalg.lstsq(X, y, rcond=None)[0]
    return X @ theta


def exp_linear_fit(x, y):
    if np.isnan(y).any():
        return 0.0 * x
    x = np.asarray(x)
    y = np.asarray(y)

    def func(x, a, k, m, b):
        return a * np.exp(-k * x) - m * x + b

    def jac(x, a, k, m, b):
        return np.column_stack(
            [
                np.exp(-k * x),
                -a * x * np.exp(-k * x),
                -x,
                np.ones_like(x),
            ]
        )

    m, b, _, _, _ = scipy.stats.linregress(x, np.log(y - y.min() + 1e-5))
    theta, _ = scipy.optimize.curve_fit(
        func,
        x,
        y,
        p0=(max(y[0] - y[-1], 0), 2e-4, max((y[-5] - y[-1]) / 10000, 0), y[0]),
        bounds=((max((y[0] - y[-1]) / 2, 0), 0.0, 0.0, 0.0), (np.inf, 0.1, np.inf, np.inf)),
        jac=jac,
    )
    return func(x, *theta)



from __future__ import annotations
from typing import Callable
import functools
import itertools
import math

from plotly.subplots import make_subplots
import plotly
import plotly.graph_objects as go
import xarray as xr

import magnify.utils as utils


def ndplot(
    xp,
    plot_function,
    facet_row: str | None = None,
    facet_col: str | None = None,
    facet_col_wrap: int = 0,
    animation_frame: str | None = None,
    **kwargs,
):
    def get_facet(idx=None):
        if animation_frame is not None:
            if idx is None:
                sub_xp = xp.isel({animation_frame: 0})
            else:
                sub_xp = xp.sel({animation_frame: idx})
        else:
            sub_xp = xp

        indexes = []
        if facet_row is not None and facet_col is not None:
            fig = make_subplots(
                rows=len(sub_xp[facet_row]),
                cols=len(sub_xp[facet_col]),
                column_titles=sub_xp[facet_col].values.astype(str),
                row_titles=sub_xp[facet_row].values.astype(str),
            )
            for i in range(len(sub_xp[facet_row])):
                for j in range(len(sub_xp[facet_col])):
                    traces = plot_function(sub_xp.isel({facet_row: i, facet_col: j}), **kwargs)
                    for trace in traces:
                        fig.add_trace(
                            trace,
                            row=i + 1,
                            col=j + 1,
                        )
        elif facet_col is not None:
            if facet_col_wrap == 0:
                num_cols = len(sub_xp[facet_col])
            else:
                num_cols = facet_col_wrap
            num_rows = math.ceil(len(sub_xp[facet_col]) / num_cols)
            fig = make_subplots(
                rows=num_rows, cols=num_cols, subplot_titles=sub_xp[facet_col].values.astype(str)
            )
            for i in range(len(sub_xp[facet_col])):
                traces = plot_function(sub_xp.isel({facet_col: i}), **kwargs)
                for trace in traces:
                    fig.add_trace(trace, row=i // num_cols + 1, col=i % num_cols + 1)
        elif facet_row is not None:
            fig = make_subplots(
                rows=len(sub_xp[facet_row]),
                cols=1,
                subplot_titles=sub_xp[facet_row].values.astype(str),
            )
            for i in range(len(sub_xp[facet_row])):
                traces = plot_function(sub_xp.isel({facet_row: i}), **kwargs)
                for trace in traces:
                    fig.add_trace(trace, row=i + 1, col=1)
        else:
            fig = go.Figure(data=plot_function(sub_xp, **kwargs))

        return fig

    fig = get_facet()
    if animation_frame is not None:
        fig.frames = [
            go.Frame(name=str(idx), data=get_facet(idx).data[0], traces=[0])
            for idx in xp[animation_frame].values
        ]
        fig.layout.sliders = [
            go.layout.Slider(
                active=0,
                steps=[
                    {
                        "args": [
                            [f.name],
                            {
                                "frame": {"duration": 0, "redraw": True},
                                "mode": "immediate",
                                "fromcurrent": True,
                                "transition": {"duration": 300, "easing": "linear"},
                            },
                        ],
                        "label": f.name,
                        "method": "animate",
                    }
                    for k, f in enumerate(fig.frames)
                ],
            )
        ]

    return fig



from __future__ import annotations
from collections.abc import Iterator
import collections
import datetime
import fnmatch
import functools
import glob
import os
import re

from numpy.typing import ArrayLike
from typing import Iterator
import bs4
import dask.array as da
import numpy as np
import pandas as pd
import tifffile
import xarray as xr

from magnify.pipeline import Pipeline
import magnify.registry as registry
import magnify.utils as utils


class Reader:
    def __init__(self) -> None:
        pass

    def __call__(
        self,
        data: ArrayLike | str,
        times: Sequence[int] | None = None,
        channels: Sequence[str] | None = None,
    ) -> Iterator[xr.Dataset]:
        if isinstance(data, str):
            data = [data]

        for d in data:
            path_dict, meta_dict = extract_paths(
                d, assay="str", channel="str", time="time", row="int", col="int"
            )
            if len(path_dict) == 0:
                raise FileNotFoundError(f"The pattern {d} did not lead to any files.")

            # None assay names should just mean a nameless assay.
            path_dict = {("",) + k[1:] if k[0] is None else k: v for k, v in path_dict.items()}
            assay_names = set(list(zip(*path_dict.keys()))[0])

            for assay_name in sorted(assay_names, key=utils.natural_sort_key):
                # Extract the paths for this assay and turn all Nones into -1 so sorting works.
                assay_dict = {
                    tuple(-1 if x is None else x for x in k[1:]): v
                    for k, v in path_dict.items()
                    if k[0] == assay_name
                }
                # Use these variables within the loop so we don't affect other assays.
                channel_coords = channels
                time_coords = times

                # Get the indices specified in the path each in sorted order.
                channel_idxs, time_idxs, row_idxs, col_idxs = (
                    sorted(set(idx)) for idx in zip(*assay_dict.keys())
                )

                # Check which dimensions are specified in the path and compute the shape of those dimensions.
                dims_in_path = []
                outer_shape = tuple()
                if channel_idxs[0] != -1:
                    dims_in_path.append("channel")
                    outer_shape += (len(channel_idxs),)
                if time_idxs[0] != -1:
                    dims_in_path.append("time")
                    outer_shape += (len(time_idxs),)
                if row_idxs[0] != -1:
                    dims_in_path.append("tile_row")
                    outer_shape += (len(row_idxs),)
                if col_idxs[0] != -1:
                    dims_in_path.append("tile_col")
                    outer_shape += (len(col_idxs),)

                # If the user didn't specify times or channels use the ones from the path.
                if time_coords is None and "time" in dims_in_path:
                    time_coords = time_idxs
                if channel_coords is None and "channel" in dims_in_path:
                    channel_coords = channel_idxs

                # Read in a single image to get the metadata stored within the file.
                with tifffile.TiffFile(next(iter(assay_dict.values()))) as tif:
                    dtype = tif.series[0].dtype
                    inner_shape = tif.series[0].shape
                    page_shape = tif.pages[0].shape

                    letter_to_dim = {
                        "C": "channel",
                        "T": "time",
                        "Z": "depth",
                        "Y": "tile_y",
                        "X": "tile_x",
                        "R": "tile_pos",
                    }
                    dims_in_file = [letter_to_dim[c] for c in tif.series[0].axes]

                    if (
                        time_coords is None
                        and tif.is_micromanager
                        and "StartTime" in tif.micromanager_metadata["Summary"]
                    ):
                        # Get the time string without timezone info.
                        time_str = tif.micromanager_metadata["Summary"]["StartTime"][:-6]
                        start_time = datetime.datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S.%f")
                        if "time" in dims_in_file:
                            # Only look at the first file's description since time isn't across multiple files.
                            planes = [
                                pl
                                for pl in bs4.BeautifulSoup(tif.pages[0].description, "xml")
                                .find("Image")
                                .find_all("Plane")
                            ]
                            assert all(pl.get("DeltaTUnit") == "ms" for pl in planes)
                            time_coords = [
                                start_time
                                + datetime.timedelta(milliseconds=float(pl.get("DeltaT")))
                                for pl in planes
                            ]
                            if "channel" in dims_in_file:
                                stride = inner_shape[dims_in_file.index("channel")]
                            else:
                                stride = 1
                            assert len(time_coords) % stride == 0
                            time_coords = time_coords[::stride]
                        else:
                            time_coords = [start_time]
                    if channel_coords is None and tif.is_micromanager:
                        channel_coords = tif.micromanager_metadata["Summary"]["ChNames"]

                    if "tile_pos" in dims_in_file:
                        # Tiles are always saved in multiple files so ignore this dimension
                        # since the user should specify tiles in their search path.
                        tile_idx = dims_in_file.index("tile_pos")
                        inner_shape = inner_shape[:tile_idx] + inner_shape[tile_idx + 1 :]
                        dims_in_file = dims_in_file[:tile_idx] + dims_in_file[tile_idx + 1 :]

                    if "depth" in dims_in_file:
                        raise ValueError("tiff files with a Z dimension are not yet supported.")
                    if "tile_y" not in dims_in_file or "tile_x" not in dims_in_file:
                        raise ValueError("tiff files must contain an X and Y dimension.")

                # Check the dimensions specified in the path and inside the tiff file do not overlap.
                if set(dims_in_file).intersection(dims_in_path):
                    raise ValueError(
                        "Dimensions specified in the path names and inside the tiff file overlap."
                    )

                # Setup a dask array to lazyload image tiles and extract attributes.
                filenames = [x for _, x in sorted(assay_dict.items())]

                def read_tile(block_id, filenames):
                    outer_id = block_id[: len(outer_shape)]
                    inner_id = block_id[len(outer_shape) :]
                    if len(outer_id) > 0:
                        file_idx = np.ravel_multi_index(outer_id, outer_shape)
                    else:
                        # This is the case where we don't have indices outside the file.
                        file_idx = 0

                    with tifffile.TiffFile(filenames[file_idx]) as tif:
                        page_ndim = len(page_shape)
                        if len(inner_shape) > page_ndim:
                            page_idx = np.ravel_multi_index(
                                inner_id[:-page_ndim], inner_shape[:-page_ndim]
                            )
                        else:
                            # This is the case where no dimensions correspond to page dims.
                            page_idx = 0
                        # Read the page from disk.
                        page = tif.pages[page_idx].asarray()
                        # Expand the dimensions of the block to incorporate outer dims and page index dims.
                        return np.expand_dims(page, axis=tuple(range(len(block_id) - page_ndim)))

                # Chunk the images so that each chunk represents a single page in the tiff file.
                tiles = da.map_blocks(
                    functools.partial(read_tile, filenames=filenames),
                    dtype=dtype,
                    chunks=(
                        (
                            tuple((1,) * size for size in outer_shape)
                            + tuple((1,) * size for size in inner_shape[: -len(page_shape)])
                            + inner_shape[-len(page_shape) :]
                        )
                    ),
                )

                coords = {}
                # Set named coordinates for channels and times if they're available.
                if channel_coords is not None:
                    coords["channel"] = channel_coords
                if time_coords is not None:
                    # Store time as seconds.
                    coords["time"] = [int(t.timestamp()) for t in time_coords]

                # Put all our data into an xarray dataset.
                assay = xr.Dataset(
                    {"tile": (dims_in_path + dims_in_file, tiles)},
                    coords=coords,
                    attrs={"name": assay_name},
                )

                # Make sure the assay always has a time and channel dimension.
                if "channel" not in assay.dims:
                    assay = assay.expand_dims("channel", 0)
                if "time" not in assay.dims:
                    assay = assay.expand_dims("time", 1)

                # Reorder the dimensions so they're always consistent and add missing dimensions.
                desired_order = ["channel", "time", "tile_row", "tile_col", "tile_y", "tile_x"]
                for dim in desired_order:
                    if dim not in assay.tile.dims:
                        assay["tile"] = assay.tile.expand_dims(dim)

                assay = assay.transpose(*desired_order)

                # Add any metadata coordinates specified by the user.
                for (meta_name, dim), meta_idxs_dict in meta_dict.items():
                    if dim == "time":
                        # Make sure we're indexing using datetimes.
                        dim_idxs = [
                            datetime.datetime.fromtimestamp(idx) for idx in assay[dim].values
                        ]
                    else:
                        dim_idxs = assay[dim].values

                    meta_idxs = [meta_idxs_dict[dim_idx] for dim_idx in dim_idxs]
                    assay = assay.assign_coords({meta_name: (dim, meta_idxs)})

                # Make sure the time dimension starts at 0 seconds.
                assay = assay.assign_coords(time=assay.time - assay.time[0])

                yield assay

    @registry.readers.register("read")
    def make():
        return Reader()


def extract_paths(pattern, **kwargs) -> dict[tuple[int, str, int, int], str]:
    default_formatters = {
        "": lambda x, y: x,
        "str": lambda x, y: x,
        "time": lambda x, y: datetime.datetime.strptime(x, y if y else "%Y%m%d-%H%M%S"),
        "int": lambda x, y: int(x),
        "float": lambda x, y: float(x),
    }

    keys = kwargs
    # Make sure keys is always a dict whose values are formatting functions.
    if not isinstance(keys, dict):
        keys = {key: "str" for key in keys}
    keys = {k: f if callable(f) else default_formatters[f] for k, f in keys.items()}
    all_keys = list(keys)

    # Format the pattern so we can do a glob search and so we can use it to extract
    # filename metadata using regexes.
    pattern = os.path.expanduser(pattern)
    meta = collections.defaultdict(dict)
    glob_path = pattern
    regex_path = fnmatch.translate(pattern)
    for key, formatter in list(keys.items()):
        # For globs replace all named search patterns e.g.: (time|%S) with the wildcard *.
        glob_path = re.sub(rf"\({key}.*?\)", "*", glob_path)
        glob_path = re.sub(rf"\([^\(]*?_{key}.*?\)", "*", glob_path)
        # For regexes replace all named search patterns with named wildcard groups.
        regex_path = re.sub(rf"\\\({key}.*?\\\)", rf"(?P<{key}>[^/\\\]*?)", regex_path)
        regex_path = re.sub(rf"\\\(([^\(]*?)_{key}.*?\\\)", r"(?P<\1>[^/\\\]*?)", regex_path)
        # Get any associated formatting information in the named search pattern.
        key_search = re.search(rf"\({key}(?:\s*\|\s*(.*?))?\)", pattern)
        if key_search:
            format_str = key_search.group(1)
            # Rebind the function, we need to set default argument values because of closure rules.
            keys[key] = lambda x, y=format_str, f=formatter: f(x, y)
        else:
            # Remove keys that weren't specified in the pattern.
            del keys[key]

        # Get meta information that provides alternate value mappings for a given key
        # along with any formatting information e.g. (concentration_time|float).
        meta_search = re.findall(
            rf"\(([^\(]*?)_{key}(?:\s*\|\s*(.*?))?(?:\s*\|\s*(.*?))?\)", pattern
        )
        for name, formatter_str, format_str in meta_search:
            meta_formatter = default_formatters[formatter_str]
            # Once again we need to set default arguments because of closure rules in Python.
            meta_formatter = lambda x, y=format_str, f=meta_formatter: f(x, y)
            meta[key][name] = meta_formatter

    regex_path = re.compile(regex_path, re.IGNORECASE)
    # Search for files matching the pattern.
    paths = glob.glob(glob_path, recursive=True)

    path_dict = {}
    meta_dict = collections.defaultdict(dict)
    for path in paths:
        match = regex_path.fullmatch(path)
        idxs = []
        for key in all_keys:
            if key in keys:
                # First get the value for the given key for the current path.
                idx = match.group(key)
                formatter = keys[key]
                idx = formatter(idx)
                idxs.append(idx)
                # Then get all the associated metadata for that key.
                for name, formatter in meta[key].items():
                    meta_idx = match.group(name)
                    meta_dict[name, key][idx] = formatter(meta_idx)
            else:
                # If the key wasn't specified in the pattern we still include it in the idxs.
                idxs.append(None)

        # Keep track of which indices correspond to which paths.
        idxs = tuple(idxs)
        if idxs not in path_dict:
            path_dict[idxs] = os.path.abspath(path)
        else:
            raise ValueError(f"{path} and {path_dict[idxs]} map to the same index.")

    return path_dict, meta_dict



from __future__ import annotations


@xr.register_dataset_accessor("mg")
class MagnifyAccessor:
    def __init__(self, xp):
        self.xp = xp

    @property
    def center(self):
        """Return the geographic center point of this dataset."""
        if self._center is None:
            # we can use a cache on our accessor objects, because accessors
            # themselves are cached on instances that access them.
            lon = self._obj.latitude
            lat = self._obj.longitude
            self._center = (float(lon.mean()), float(lat.mean()))
        return self._center

    def save(self):
        """Plot data on a map."""
        return "plotting!"



from __future__ import annotations
from typing import Callable, Sequence

import confection
from numpy.typing import ArrayLike
import xarray as xr

import magnify.registry as registry
import magnify.utils as utils


class Pipeline:
    def __init__(self, reader: str, config: dict[str, dict[str, str]]):
        self.reader: Callable[[ArrayLike | str], xr.Dataset] = registry.readers.get(reader)()
        self.config = confection.Config(config)
        self.components: list[Callable[[xr.Dataset], xr.Dataset]] = []

    def __call__(
        self,
        data: ArrayLike | str,
        times: Sequence[int] | None = None,
        channels: Sequence[str] | None = None,
    ) -> xr.Dataset | list[xr.Dataset]:
        inputs = self.reader(data=data, times=times, channels=channels)
        assays = []
        for assay in inputs:
            for name, component in self.components:
                assay = component(assay)

            assays.append(assay)

        if len(assays) == 1:
            assays = assays[0]

        return assays

    def add_pipe(
        self,
        name: str,
        after: str | int | None = None,
        before: str | int | None = None,
        first: bool = False,
        last: bool = False,
    ) -> None:
        component_factory = registry.components.get(name)
        component = component_factory(**utils.valid_kwargs(self.config, component_factory))

        if after is None and before is None and not first and not last:
            last = True
        if (after is not None) + (before is not None) + first + last > 1:
            raise ValueError("Only one of after, before, first, and last can be set.")
        new_element = (name, component)
        if first:
            idx = 0
        elif last:
            idx = len(self.components)
        elif isinstance(before, int):
            idx = before
        elif isinstance(before, str):
            idx = [name for (name, _) in self.components].index(before)
        elif isinstance(after, int):
            idx = after + 1
        elif isinstance(after, str):
            idx = [name for (name, _) in self.components].index(after) + 1

        self.components.insert(idx, new_element)
